open Pcaml;;

(* rip-off from str.ml *)
let quote s =
  let len = String.length s in
  let buf = String.create (2 * len) in
  let pos = ref 0 in
  for i = 0 to len - 1 do
    match s.[i] with
      '[' | ']' | '*' | '.' | '\\' | '?' | '+' | '^' | '$' as c ->
        buf.[!pos] <- '\\'; buf.[!pos + 1] <- c; pos := !pos + 2
    | c ->
        buf.[!pos] <- c; pos := !pos + 1
  done;
  String.sub buf 0 !pos;;

type tree =
    String of string
  | Surround of tree
  | Tagged of string*tree
  | Concat of tree*tree
  | Expr of MLast.expr
  | Optional of tree

let surround = function
    (Surround x) as xs -> xs
  | x -> Surround x

let tag t = function
    Surround x -> Tagged(t, x)
  | x          -> Tagged(t, x)

let string x = String(quote x)

let ( ^. ) x y = Concat(x,y)

let regexp = Grammar.Entry.create gram "regexp" and
    regexp1 = Grammar.Entry.create gram "regexp" and
    char_set = Grammar.Entry.create gram "regexp" and
    re_match = Grammar.Entry.create gram "REmatch" and
    labeled_patt = Grammar.Entry.create gram "REmatch labeled patt" and
    fun_def = Grammar.Entry.create gram "REmatch labeled patt" and
    re_match_what_to_do = Grammar.Entry.create gram "REmatch";;

let rec flatten_expr = function
    String(s) -> String(s)
  | Surround(s)  -> flatten_expr ((String "\\(") ^. s ^. (String "\\)"))
  | Tagged(t, s) -> flatten_expr ((String "\\(") ^. s ^. (String "\\)"))
  | Optional(x) -> flatten_expr x
  | Concat(x,y) ->
      let x'=flatten_expr x and y'=flatten_expr y in
      begin
      	match x',y' with
	  (String a), (String b) -> String(a ^ b)
      	| _ -> Concat(x', y')
      end
  | Expr(e) -> Expr(e)

let rec convert_expr loc = function
    String(s) -> <:expr< $str:s$ >>
  | Concat(x,y) -> <:expr< $(convert_expr loc x)$ ^ $(convert_expr loc y)$ >>
  | Expr(e) -> e

module Marker_map = Map.Make
  (struct
     type t = string
     let compare x y = - (Pervasives.compare x y)
   end)

let retrieve_markers =
  let rec retrieve ((offset, markers) as input_data) flag loc = function
    String(_) | Expr(_) -> input_data
  | Surround(x) -> retrieve ((succ offset), markers) flag loc x
  | Concat(x, y) -> let after_data = retrieve input_data flag loc x in
                    retrieve after_data flag loc y
  | Tagged(t, x) -> retrieve
	((succ offset), (Marker_map.add t (offset, flag) markers)) flag loc x
  | Optional(x) -> retrieve input_data true loc x
  in retrieve (1, Marker_map.empty) false;;

let rec filter_matchings = function
    [] -> (None, [])
  | (None, default_clause)::_ -> ((Some default_clause), [])
  | ((Some e), clause)::tail ->
      let default, other_matchings = filter_matchings tail in
      default, ((e,clause)::other_matchings)

EXTEND
  regexp1:[[ "["; x=char_set; "]"      -> String ("[" ^ x ^ "]")
	   | "["; "^"; x=char_set; "]" -> String ("[^" ^ x ^ "]")

           | "("; x=regexp; ")"        -> surround x
	   | t=LIDENT; ":"; "("; x=regexp; ")" -> tag t x

           | x=regexp1; "+"            -> (surround x) ^. (String "+")
	   | x=regexp1; "*"            -> Optional
		                            ((surround x) ^. (String "*"))
	   | x=regexp1; "?"            -> Optional
		                            ((surround x) ^. (String "?"))

	   | "{"; e=expr ;"}"          -> Expr e
	   | x=LIDENT                  -> Expr <:expr< $lid:x$ >>

           | x=STRING                  -> string x
	   | x=CHAR                    -> string x ]];

  regexp: [[ x=regexp1                 -> x
           | x=regexp; "|"; y=regexp   -> x ^. (String "\\|") ^. y
           | x=regexp; y=regexp        -> x ^. y ]];

  char_set: [[ x=CHAR; "-"; y=CHAR     -> x ^ "-" ^ y
             | x=char_set ; y=char_set -> x ^ y
             | x=CHAR                  -> x (* WRONG QUOTING *) ]];

  expr:  LEVEL "simple"
     [[ "RE"; "{"; x=regexp ; "}" ->
           let regexp_string = convert_expr loc (flatten_expr x) and
	       (nparen, markers) = retrieve_markers loc x in
	   let extractor =
	     let right_hand_side =
	       Marker_map.fold
	         begin
		   fun label (offset, is_optional) previous ->
		     Printf.fprintf stderr "%s => %d\n" label offset;
		     let ofs = string_of_int offset in
		     if is_optional
		     then
		       <:expr< $previous$ ~$label$:
		       try Some (Str.matched_group $int:ofs$ str)
		       with [Not_found -> None] >> 
		     else
		       <:expr< $previous$ ~$label$:
		       (Str.matched_group $int:ofs$ str) >> 

		   (* ((Str.group_beginning $int:ofs$),
		      (Str.group_end $int:ofs$)) *)
		 end
		 markers
		 ~init: <:expr< what_to_do
                            ~contents: (Str.matched_string str) >>
	     in <:expr< fun str what_to_do -> $right_hand_side$ >>
	   and regexp_compiled = <:expr< Str.regexp $regexp_string$>>
	   in <:expr< ( $list: [regexp_compiled; extractor]$ ) >> ]];

  expr:  LEVEL "expr1"
     [[ "REmatch"; s=expr; "with" ; OPT "|";
         all_matchings = LIST1 re_match SEP "|" ->
	   let default, matchings = filter_matchings all_matchings in
	   let actual_work =
             List.fold_right
               begin
		 fun (regexp_expr, what_to_do) other_cases ->
		   <:expr<
                    let (__regexp_rxp, __regexp_extractor) = $regexp_expr$ in
                     try
                       do ignore
                         (Str.search_forward __regexp_rxp __regexp_what 0);
                       return __regexp_extractor __regexp_what $what_to_do$
                     with
                     [ Not_found -> $other_cases$ ] >>
               end
               matchings
	       (match default with
		 None -> <:expr< () >>
	       | Some e -> e)
	   in <:expr< let __regexp_what = $s$ in $actual_work$ >> ]];

  re_match: [[ e=expr; what_to_do=re_match_what_to_do
	         -> ((Some e), what_to_do)
             | "_" ; "->"; what_to_do=expr -> (None, what_to_do) ]];

  labeled_patt:
    [ [ i = TILDEIDENTCOLON; p = patt LEVEL "simple" ->
          <:patt< ~ $i$ : $p$ >>
      | i = TILDEIDENT ->
          <:patt< ~ $i$ >>
      | i = QUESTIONIDENTCOLON; j = LIDENT ->
          <:patt< ? $i$ : $lid:j$ >>
      | i = QUESTIONIDENTCOLON; "("; j = LIDENT; "="; e = expr; ")" ->
          <:patt< ? $i$ : ( $lid:j$ = $e$ ) >>
      | i = QUESTIONIDENT ->
          <:patt< ? $i$ : $lid:i$ >>
      | "?"; "("; i = LIDENT; "="; e = expr; ")" ->
          <:patt< ? ( $i$ = $e$ ) >> ]];

  fun_def:
    [ RIGHTA
      [ p = patt LEVEL "simple"; e = fun_def -> <:expr< fun $p$ -> $e$ >>
      | "->"; e = expr                       -> <:expr< $e$ >>
      |	p = labeled_patt; e = fun_def        -> <:expr< fun $p$ -> $e$ >> ]];

  re_match_what_to_do :
    [[ "->"; what_to_do=expr -> what_to_do
     | "as"; p = patt LEVEL "simple"; e = fun_def ->
          <:expr< fun $p$ -> $e$ >>
     | "as"; p = labeled_patt; e = fun_def ->
          <:expr< fun $p$ -> $e$ >>]];
END;;
